#! /usr/bin/python

import os, sys
import re
import util
import operator
import numbers
import random
import numpy as np
import math
from collections import defaultdict
from collections import OrderedDict



class Sentences:
    def __init__(self, gfile, tagdict):
        self.dict = defaultdict(list)
        self.mapping = {}
        self.tagdict = tagdict
        self.headdict = {} 
        self.feedback = defaultdict(list)
        self.feedback['deprel'] = defaultdict(list)
        self.feedback['attach'] = defaultdict(list)
        self.processed = {}
        c = 0
        n = 0
        t = 0
        goldlist = util.load_list_from_file(gfile)
        self.feedback[c] = []
        for g in goldlist: 
                if g == '\n' or g == '':
                    c += 1
                    t = 0 
                    self.feedback['deprel'][c] = []
                    self.feedback['attach'][c] = []
                else: 
                    tid, wrd, lem, tag, x, y, head, dep, a, b = g.split("\t") 
                    tid = int(tid)-1
                    head = int(head)-1
                    arc = str(tid)+":"+str(head)+":"+dep 
                    self.feedback['deprel'][c].append(None)
                    self.feedback['attach'][c].append(None)
                    self.dict[c].append({'w':wrd,'g_tag':tag, 'g_rel':arc,'a':[], 'l':[]})
                    self.mapping[n] = {}
                    self.mapping[n]['sid'] = c
                    self.mapping[n]['tid'] = t
                    # here we need a combination of head and dep
                    n += 1
                    t += 1


        
    # take a list of dictionaries for a sentence
    # print it to Stdout and collect the manual
    # annotation for the token at position n
    # save the feedback
        
    def getAnnotation(self, ind):

        sid = self.mapping[ind]['sid']
        tid = self.mapping[ind]['tid']
        annots = {}
        for annot in self.dict[sid][tid]['a']:
            annots[int(annot.split(':')[2])] = 1

        print
        print ("\nGet new annotation for sentence ", sid+1, ",\ttoken ", (tid+1), ":\t<\033[1;34m", self.dict[sid][tid]['w'], "\033[1;m>\n")
        print

        for n in range(len(self.dict[sid])): 
            if tid == n:
                print ('\033[1;34m', self.dict[sid][n]['w'], '\033[1;m', end=" ")
            elif n in annots: 
                print ('\033[1;31m'+ self.dict[sid][n]['w']+ '_'+ str(n)+ '\033[1;m', end=" ")
            else:
                print (self.dict[sid][n]['w'], end=" ")
        print
        print ("\n\nPredicted label(s): ", list(OrderedDict.fromkeys(self.dict[sid][tid]['l']))) 
        # now read the dep label from stdin and return it 
        oracle_annot = input("\nPlease enter the correct head and label\n(or type: <stop now> to end program)\n ")
        while len(oracle_annot.split(' ')) < 2:
               oracle_annot = input("\nPlease enter head AND label: ")	
        head, dep = oracle_annot.split(' ')
 
        print 
        # sanity check: is the label included in our label dictionary?
        while (self.checkTag(dep) == False):
            if dep == "now" and head == "stop":
                 print("RETURN STOP")
                 return sid, tid, dep, head
            print
            print (dep, " is not a valid label! Please try again...\n")
            oracle_annot = input("Enter the correct head and label: ")
            while len(oracle_annot.split(' ')) < 2:
                 oracle_annot = input("\nPlease enter head AND label: ")
            head, dep = oracle_annot.split(' ')
 
        dep = dep.upper()
 
        # store feedback
        self.feedback['deprel'][sid][tid] = dep 
        self.feedback['attach'][sid][tid] = head 
        return sid, tid, dep, head


    # for AL simulations: get the correct label from the gold standard
    def getOracleTag(self, ind):

        sid = self.mapping[ind]['sid']
        tid = self.mapping[ind]['tid']
        arc = self.dict[sid][tid]['g_rel']
        dep = arc.split(':')[2]
        head= arc.split(':')[1] 
        self.feedback['deprel'][sid][tid] = dep
        self.feedback['attach'][sid][tid] = arc.split(':')[1]
        return sid, tid, dep, head


    # check if tag is a valid tag from the tagset
    def checkTag(self, tag): 
        if tag in self.tagdict:
            return True
        elif tag.upper() in self.tagdict:
            return True
        else:
            return False
    

    # take a table with predictions (from the annotators/parsers)
    # and add them to the sentences object
    def addPredictions(self, table): 
        cols = len(table['arc'])
        rows = len(table['arc'][0])
        sid  = 0
        tid  = -1 
        # self.mapping: 
        # {0: {'tid': 0, 'sid': 1}, 1: {'tid': 1, 'sid': 1}, 2: {'tid': 2, 'sid': 1}, ...
        
        # first read in all predictions
        for r in range(rows):    

            # get sid for this row index     
            if (self.mapping[r]['sid'] > sid):
                tid = -1 
            sid = self.mapping[r]['sid']
            tid += 1 
            
            for c in range(cols): 
                self.dict[sid][tid]['l'].append(table['arc'][c][r].split(':')[3])
                self.dict[sid][tid]['a'].append(table['arc'][c][r]) 

        # next, add headdict
        sid  = 0
        tid  = -1 
        for r in range(rows):                    
            # get sid for this row index     
            if (self.mapping[r]['sid'] > sid):
                tid = -1 
            sid = self.mapping[r]['sid']
            tid += 1 
            
            for c in range(cols):            
                for i in range(len(self.dict[sid][tid]['a'])):
                    headtag = "" 
                    s, t, hd, dep, depPos = self.dict[sid][tid]['a'][i].split(':')                      
                    s = int(s); t = int(t); hd = int(hd)
                    headPos =  self.dict[s][hd]['a'][i].split(':')[4]
                    dir = 'l'
                    if t < hd:
                        dir = 'r'
                    headtag = depPos + dir + headPos
                    if headtag not in self.headdict:
                        new = len(self.headdict) +1
                        self.headdict[headtag] = new   
        
 


    def getSentByInd(self, ind):      
        sid = self.mapping[ind]['sid']
        return self.dict[sid]


    def printSent(self, n):
        for ind in range(len(self.dict[n])):
            print (ind, " ", self.dict[n][ind]['w'])


    def getGoldTag(n):
        sid = self.mapping[n]['sid']
        tid = self.mapping[n]['tid']

     

    # write feedback file with corrected annotations
    # (either from oracle or from goldstandard/AL simulation)    
    # and combine them with predictions from MACE

    def write_feedback_to_file(self, file_path, newAnnotator, c):
        try:
            file_out = open(file_path, "w") 
            for sid in self.dict.keys():
                for row in range(len(self.dict[sid])):
                    if c == 1:
                        if self.feedback['deprel'][sid][row] == None:
                            file_out.write("\n")
                        else:
                            file_out.write(str(self.tagdict[self.feedback['deprel'][sid][row]])+"\n")
                    else:
                        if self.feedback['attach'][sid][row] == None:
                            file_out.write("\n")
                        else:
                            file_out.write(str(self.feedback['attach'][sid][row])+"\n")

            file_out.close()
        except IOError as e:
            print ("I/O error({0}): {1}".format(e.errno, e.strerror))



    def update_predictions(self, newAnnotator, c):
        for sid in self.dict.keys():
                for row in range(len(self.dict[sid])):
                    sentId, tokId, headId, deprel, pos = self.dict[sid][row]['a'][newAnnotator].split(':')
                    # update randomly chosen classifier     
                    if self.feedback['deprel'][sid][row] != None:   
                        new =  sentId + ":" + tokId + ":" + self.feedback['attach'][sid][row] + ":" + self.feedback['deprel'][sid][row] + ":" + pos
                        self.dict[sid][row]['a'][newAnnotator] = new
 



    def backupFiles(self, c):
        
        # back-up the old preds file
        command = "mv preds.csv preds.csv." + str(c)
        os.system(command)

        # and the feedback file as well        
        if c%2 == 1: 
            command = "cp feedback.attach feedback." + str(c)
        else:
            command = "cp feedback.deprel feedback." + str(c)
        os.system(command)

        # print table content to csv file
        self.printCSV()

        


    # generate csv file, print table to file
    def printCSV(self):
        
        with open('preds-deprel.csv', 'w') as csv_file:
            for sid in self.dict.keys():
                for row in range(len(self.dict[sid])):
                    string = ""
                    for pred in range(len(self.dict[sid][row]['a'])):
                        tag = self.dict[sid][row]['a'][pred].split(':')[3]     

                        if util.number(tag):
                            string += str(tag) + "," 
                        else:                        
                            string += str(self.tagdict[tag]) + ","
                            
                    string = string[:-1]
                    string += '\n'
                    csv_file.write(string) 

        with open('preds-attach.csv', 'w') as csv_file:
            for sid in self.dict.keys():
                for row in range(len(self.dict[sid])):
                    string = ""
                    for pred in range(len(self.dict[sid][row]['a'])):
                        tag = self.dict[sid][row]['a'][pred].split(':')[2]   
                        string += str(tag) + "," 
                            
                    string = string[:-1]
                    string += '\n'
                    csv_file.write(string)    

                    


    #####
    # take two lists and compute acc. for list elements
    def eval_labels(self, gold):
        cols = len(self.dict[1][0]['a'])
        result = '\nAccuracy for individual annotators:\n\n'
        for c in range(cols):
            preds = []
            total = 0
            correct = 0
            for sid in self.dict.keys():
                for tok in range(len(self.dict[sid])):
                    preds.append(self.dict[sid][tok]['a'][c])

            # sanity check: are lists of same length?
            if len(gold) != len(preds):
                print ("gold and pred lists are of different length\n", file=sys.stderr)
                print (len(gold), " - ", len(preds), "\n", file=sys.stderr)
                sys.exit(1)
            
            for i in range(0, len(gold)):
                if preds[i] != '':  
                    pred_i = preds[i].split(':')
                    eval = pred_i[0] +':'+ pred_i[1] +':'+ pred_i[2] +':'+ pred_i[3]
                    if gold[i] == eval:
                        correct +=1
                    total +=1

            result += "\t\tLAS: " + str(float(correct)/total*100) + "\t(" + str(correct) + "/" + str(total) + ")\n"
        return result



    # compute accuracy for majority vote baseline
    def majority_baseline(self, gold):
    
        cols = len(self.dict[1][0]['a'])
        correct = 0
        total = 0

        for sid in self.dict.keys():
            for tok in range(len(self.dict[sid])):
                items = self.dict[sid][tok]['a']

                counts = [ (items.count(x), x) for x in set(items) ]
                highest = max(counts)[0]
                candidates = [count[1] for count in counts if count[0] == highest]
                truth = random.choice(candidates).split(':')
                eval = truth[0] +":"+ truth[1] +":"+ truth[2] +":"+ truth[3]
                # compare vote to gold tag
                if gold[total] == eval: 
                    correct +=1
                total += 1
                
        return "\nMAJORITY VOTE\tLAS: " + str(float(correct)/total*100) + "\t(" + str(correct) + "/" + str(total) + ")\n\n"


        
        
    # get the n votes with the highest entropy (based on entropy file    
    # generated by MACE) for manual annotation / AL simulation
    def entropy(self, n):
        entropies = util.load_dict_from_file( 'entropies' )
        indices = []
        stop = 0
        for x in sorted(entropies, key=float, reverse=True):
            if entropies[x] in self.processed:
                continue
            indices.append(entropies[x]-1)
            self.processed[entropies[x]] = 1
            stop += 1
            if stop == n:
                return indices
      

        

    # to be used for majority baseline
    # compute entropy based on each row in the annotations
    def get_entropy_for_row(self, N):
        cols = len(self.dict[1][0]['a'])
        entropies = {}
        c = 0
        
        for sid in self.dict.keys():
            for tok in range(len(self.dict[sid])):
                row = self.dict[sid][tok]['a']
                list = [row.count(x) for x in set(row)]
                sum = 0
                for i in list:
                    x = float(i)/float(7)
                    sum += (x * math.log(i))
                entropies[c] = sum
                c += 1

        indices = []
        stop = 0
        for x in sorted(entropies, key=entropies.__getitem__, reverse=False):
                       
            if x in self.processed:
                continue
            indices.append(x)
            self.processed[x] = 1
            stop += 1
            if stop == N:
                return indices
            
 
