#coding:utf-8
import string
from multiprocessing import Process
from random import shuffle
import collections
import operator
from collections import Counter
import numpy as np
from dependency_decoding import chu_liu_edmonds 
import re
import os, sys
import operator
import random
import numbers
import math 


def selectSent(table, list):
    print (list)
    

def get_entropy(row):
    list = [row.count(x) for x in set(row)]
    sum = 0
    for i in list:
        x = float(i)/float(7)
        sum += (x * math.log(i))
    return sum

       
def load_list_from_file(list_file_path):
    list_file = open(list_file_path)
    list = [line.strip() for line in list_file]
    return list


def load_set_from_file(list_file_path):
    list_file = open(list_file_path)
    list = [line.strip() for line in list_file]
    return set(list)


# key : index
def load_dict_from_file(dict_file_path):
    dict = {}
    dict_file = open(dict_file_path)
    lines = [line.strip() for line in dict_file]
    for index, line in enumerate(lines):
        if line == "":
            continue
        dict[line] = index+1
    dict_file.close()
    return dict


# key : index
def load_mapping_from_file(dict_file_path):
    dict = {}
    dict_file = open(dict_file_path)
    for line in dict_file:
        if line.strip() == "":
            continue
        ftag, ctag = line.split("\t")
        dict[ftag] = ctag.strip()
    dict_file.close()
    return dict


# write dict keys to file
def write_dict_keys_to_file(dict, file_path):
    file_out = open(file_path,"w")
    file_out.write("\n".join([str(key) for key in sorted(dict, key=lambda i: int(dict[i]))]))
    file_out.close()


# read all files in path ending with suffix
# append file context to table and return table
def readPredictions( path, suffix ):
        files = os.listdir( path )
        ind = 0
        table = { 
            'arc': [],
            'l': [],
            'a': []
        }        
        filelist = []

        # get all files in path with suffix
        # sort file list alphabetically
        files = sorted(files)
        
        for file in files:
            print (file)
            if file.split(".")[1] == suffix:
                filelist.append(file)

        # read all files, split the lines and store the tags in table
        for file in filelist:
                file = path + "/" + file
                print ("... reading file " + file + " ...")
        
                with open(file) as f:
                        table['arc'].append(ind)
                        table['a'].append(ind)
                        table['l'].append(ind)
                        sid = 0 
                        arcs = []
                        labels = []
                        heads = []
                        for line in f:
                                if line == '\n':
                                    sid += 1
                                else:
                                    tid, wrd, lem, tag, x, y, head, dep, a, b = line.split("\t")
                                    tid = int(tid) -1 # first token is 0
                                    head = int(head) -1
                                    # we now append the directed arc with its label
                                    arc = str(sid)+":"+str(tid)+":"+str(head)+":"+dep+":"+tag  
                                    arcs.append(arc) 
                                    labels.append(dep)
                                    heads.append(int(head))
                        table['arc'][ind] = arcs 
                        table['l'][ind] = labels
                        table['a'][ind] = heads
                        ind+=1
        return table



def print_log(logfile, string):
    old_stdout = sys.stdout
    logfile = open(logfile, "a")
    logfile.write(string)
    logfile.close()


def read_competence(file_path):
        f = open(file_path)
        for line in f:
            return line.strip().split('\t')
            


def print_trees(pred_deprel, pred_attach, sentences, c):

    # read competence scores and use POS tags from the first parser (pos_annot = 0)
    # merge with head and deprel predictions from MACE
    pos_annot = 0
    # we need to map the numbers for deprel labels back to their tags
    mapping = {}
    for tag in sentences.tagdict:
        mapping[ sentences.tagdict[tag] ] = tag

    #print mapping   
    filename = "./trees" + str(c) + ".txt" 
    outfile = open(filename, "w")

    n=0
    for sid in range(0, len(sentences.dict)):
        for tok in range(0, len(sentences.dict[sid])): 
            string = str(tok+1) + "\t" + sentences.dict[sid][tok]['w'] + "\t_\t"  
            string = string + sentences.dict[sid][tok]['a'][pos_annot].split(':')[4] + "\t"
            string = string + sentences.dict[sid][tok]['a'][pos_annot].split(':')[4] + "\t_\t"  
            string = string + str(int(pred_attach[n])+1) + "\t"
            string = string + str(mapping[ int(pred_deprel[n]) ]) + '\n'
 
            outfile.write(string) 
            n += 1
        outfile.write('\n') 
    outfile.close()  



def print_trees_CLE(pred_deprel, pred_attach, sentences, c):

    # read competence scores and use POS tags from the first annotator (pos_annot = 0)
    pos_annot = 0

    # first generate the best unlabelled tree
    # (also add the MACE predictions as an additional annotator)
    # and then add the labels to the unlabelled max spanning tree

    # we need to map the numbers for deprel labels back to their tags
    mapping = {}
    heads_all = []
    for tag in sentences.tagdict:
        mapping[ sentences.tagdict[tag] ] = tag
    
    #print mapping   
    filename = "./trees" + str(c) + "-cle.txt" 
    outfile = open(filename, "w")
    n=0

    for sid in range(0, len(sentences.dict)):
        # get the matrix for this sentence
        slen = len(sentences.dict[sid])
        matrix = np.zeros((slen+1, slen+1))

        for tok in range(0, len(sentences.dict[sid])): 
            # get all head candidates for this token 
            for pred in sentences.dict[sid][tok]['a']:
                 matrix[tok+1][int(pred.split(':')[2]) +1] += 1  
            n += 1  

        heads, tree_score = chu_liu_edmonds(matrix)    
        for c in range(1, len(heads)):
            heads_all.append(heads[c])

    # now get the deprel labels for all sentences
    # from the MACE predictions and add the heads from CLE
    n=0
    for sid in range(0, len(sentences.dict)):
        for tok in range(0, len(sentences.dict[sid])): 
            string = str(tok+1) + "\t" + sentences.dict[sid][tok]['w'] + "\t_\t"  
            string = string + sentences.dict[sid][tok]['a'][pos_annot].split(':')[4] + "\t"
            string = string + sentences.dict[sid][tok]['a'][pos_annot].split(':')[4] + "\t_\t"  
            string = string + str(heads_all[n]) + "\t"
            string = string + str(mapping[ int(pred_deprel[n]) ]) + '\n'
 
            outfile.write(string) 
            n += 1
        outfile.write('\n') 
    outfile.close() 




def print_trees_CLE_weighted(pred_deprel, pred_attach, sentences, c):

    # read competence scores and use POS tags from the first annotator (pos_annot = 0)
    pos_annot = 0

    # first generate the best unlabelled tree 
    # and then add the labels to the unlabelled max spanning tree
    # weight the label predictions according to the competence of the model

    # we need to map the numbers for deprel labels back to their tags
    mapping = {}
    heads_all = []
    for tag in sentences.tagdict:
        mapping[ sentences.tagdict[tag] ] = tag

    # read competence (from MACE)
    competence = read_competence('competence')

    #print mapping   
    filename = "./trees" + str(c) + "-cle-w.txt" 
    outfile = open(filename, "w")
    n=0

    for sid in range(0, len(sentences.dict)):
        # get the matrix for this sentence
        slen = len(sentences.dict[sid])
        matrix = np.zeros((slen+1, slen+1))
        for tok in range(0, len(sentences.dict[sid])): 
            # get all head candidates for this token 
            a = 0 # a is for annotator
            for pred in sentences.dict[sid][tok]['a']: 
                matrix[tok+1][int(pred.split(':')[2]) +1] += float(competence[a])  
                a += 1
            n += 1  

        heads, tree_score = chu_liu_edmonds(matrix)    
        for c in range(1, len(heads)):
            heads_all.append(heads[c])

    # now get the deprel labels for all sentences
    # from the MACE predictions and add the heads from CLE
    n=0
    for sid in range(0, len(sentences.dict)):
        for tok in range(0, len(sentences.dict[sid])): 
            string = str(tok+1) + "\t" + sentences.dict[sid][tok]['w'] + "\t_\t"  
            string = string + sentences.dict[sid][tok]['a'][pos_annot].split(':')[4] + "\t"
            string = string + sentences.dict[sid][tok]['a'][pos_annot].split(':')[4] + "\t_\t"  
            string = string + str(heads_all[n]) + "\t"
            string = string + str(mapping[ int(pred_deprel[n]) ]) + '\n'
 
            outfile.write(string) 
            n += 1
        outfile.write('\n') 
    outfile.close() 




def number(x):
    if isinstance(x, numbers.Number):
        return True
    return False

